import os
from pathlib import Path

from scriptengine.exceptions import ScriptEngineTaskRunError
from scriptengine.tasks.core import Task, timed_runner

_DEFAULT_LMOD_ENVVAR = "LMOD_DIR"
_DEFAULT_ENV_MOD_PATH = Path("/usr/share/Modules")


def modfunc_from_ini_script(init_script):
    try:
        ini_script_code = init_script.read_text()
    except FileNotFoundError as e:
        raise RuntimeError(f"Init script not found: {e}")
    exec_globals = dict()
    try:
        exec(ini_script_code, exec_globals)
    except Exception as e:
        raise RuntimeError(f"Init script error: {e}")
    try:
        return exec_globals["module"]
    except KeyError:
        raise RuntimeError("Init script does not define 'module' function")


def modfunc_from_lmod(lmod_init_script=None):
    """Returns 'module' function or None
    Lmod implementation, see
    https://lmod.readthedocs.io
    https://github.com/TACC/Lmod
    """
    if lmod_init_script:
        lmod_init_script = Path(lmod_init_script)
    elif _DEFAULT_LMOD_ENVVAR in os.environ:
        lmod_init_script = (
            Path(os.environ[_DEFAULT_LMOD_ENVVAR]).parent / "init/env_modules_python.py"
        )
    else:
        return None
    return modfunc_from_ini_script(lmod_init_script)


def modfunc_from_environment_modules(env_mod_init_script=None):
    """Returns 'module' function or None
    Implementation for Environment modules, see
    https://modules.readthedocs.io
    http://modules.sourceforge.net
    """
    if env_mod_init_script:
        env_mod_init_script = Path(env_mod_init_script)
    elif _DEFAULT_ENV_MOD_PATH.exists():
        env_mod_init_script = _DEFAULT_ENV_MOD_PATH / "init/python.py"
    else:
        return None
    return modfunc_from_ini_script(env_mod_init_script)


class Module(Task):
    """hpc.module"""

    _required_arguments = ("cmd",)
    _module_func = None

    def __init__(self, arguments):
        Module.check_arguments(arguments)
        super().__init__(arguments)

    @timed_runner
    def run(self, context):
        self.log_info("Execute module command")

        cmd = self.getarg("cmd", context)
        args = self.getarg("args", context, default=[])
        self.log_debug(f"Run module cmd '{cmd}' with args {args}")

        if not Module._module_func:
            init = self.getarg("init", context, default=None)
            self.log_debug(
                f"Module system not yet initialised. Initialising from '{init}'"
            )
            try:
                Module._module_func = modfunc_from_lmod(
                    init
                ) or modfunc_from_environment_modules(init)
            except RuntimeError as e:
                self.log_error(f"Error initialising the module system: {e}")
                raise ScriptEngineTaskRunError
            if not Module._module_func:
                self.log_error("Error initialising the module system")
                raise ScriptEngineTaskRunError

        try:
            Module._module_func(cmd, *args)
        except Exception as e:
            self.log_error(f"Error while running module command: {e}")
            raise ScriptEngineTaskRunError


class ModuleLoad(Module):
    """hpc.module.load"""

    _required_arguments = ("names",)

    def __init__(self, arguments):
        arguments["cmd"] = "load"
        ModuleLoad.check_arguments(arguments)
        load_args = arguments.pop("names")
        arguments["args"] = load_args if isinstance(load_args, list) else [load_args]
        super().__init__(arguments)



"""setup.py for package scriptengine-tasks-hpc."""
import os
import codecs
import setuptools


def read(rel_path):
    """
    Helper function to read file in relative path.
    """
    here = os.path.abspath(os.path.dirname(__file__))
    with codecs.open(os.path.join(here, rel_path), 'r') as fp:
        return fp.read()


def get_version(rel_path):
    """
    Helper function to get package version.
    """
    for line in read(rel_path).splitlines():
        if line.startswith('__version__'):
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]


setuptools.setup(
    name='scriptengine-tasks-hpc',
    version=get_version('scriptengine_hpc/version.py'),
    author='Uwe Fladrich',
    author_email='uwe.fladrich@protonmail.com',
    description='ScriptEngine tasks you may need on HPC systems',
    long_description=read('README.md'),
    long_description_content_type='text/markdown',
    url='https://github.com/uwefladrich/scriptengine-tasks-hpc',
    packages=setuptools.find_packages(),
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.6',
    install_requires=[
        'scriptengine>=0.8.0',
    ],
    entry_points={
        'scriptengine.tasks': [
            'hpc.slurm.sbatch = scriptengine_hpc.slurm:Sbatch',
            'hpc.module = scriptengine_hpc.module:Module',
            'hpc.module.load = scriptengine_hpc.module:ModuleLoad',
        ],
    },
)



import os
import sys
import subprocess

from scriptengine.tasks.core import Task, timed_runner
from scriptengine.jinja import render as j2render
from scriptengine.exceptions import ScriptEngineStopException, \
                                    ScriptEngineTaskRunError


_JOB_VAR = 'SLURM_JOB_NAME'
_SBATCH_CMD = 'sbatch'


class Sbatch(Task):

    @timed_runner
    def run(self, context):

        def in_batch_job():
            return os.environ.get(_JOB_VAR) is not None

        def do_recursive_submit():
            return self.getarg(
                'submit_from_batch_job',
                context,
                default=False,
            )

        def is_sbatch_opt(opt):
            return opt not in (
                'scripts',
                'submit_from_batch_job',
                'hetjob_spec',
            ) and not opt.startswith('_')

        if in_batch_job() and not do_recursive_submit():
            self.log_debug(
                'Already running within a batch job and '
                '"submit_from_batch_job" is not set or false - '
                'not submitting new job'
            )
            return

        # Start building the sbatch command line
        sbatch_cmd_line = [_SBATCH_CMD]

        sbatch_general_args = []
        for opt, arg in self.__dict__.items():
            if is_sbatch_opt(opt):
                sbatch_general_args.append(f'--{opt}')
                if arg:
                    sbatch_general_args.append(j2render(arg, context))
        sbatch_cmd_line.extend(map(str, sbatch_general_args))

        hetjob_spec = self.getarg('hetjob_spec', context, default=[])
        if hetjob_spec:
            self.log_debug('Submitting a hetereogeneous SLURM job with SBATCH')
            sbatch_hetjob_args = []
            for job_args in hetjob_spec:
                if len(sbatch_hetjob_args) > 0:
                    sbatch_hetjob_args.append(':')
                for opt, arg in job_args.items():
                    sbatch_hetjob_args.append(f'--{opt}')
                    if arg:
                        sbatch_hetjob_args.append(j2render(arg, context))
            sbatch_cmd_line.extend(map(str, sbatch_hetjob_args))

        scripts = self.getarg('scripts', context, default=None)
        if scripts:
            sbatch_cmd_line.append('se')
            sbatch_cmd_line.extend(
                map(
                    str,
                    scripts if isinstance(scripts, list) else [scripts],
                )
            )
        else:
            # If no scripts were given, use the original SE command line
            sbatch_cmd_line.extend(sys.argv)
        self.log_debug(f'Command line for submitting job: {sbatch_cmd_line}')

        # Run sbatch command, with handling of errors
        try:
            subprocess.run(sbatch_cmd_line, check=True)
        except subprocess.CalledProcessError as e:
            self.log_error(
                f'Submit command returns error: {e}'
            )
            raise ScriptEngineTaskRunError
        else:
            self.log_info(
                'Requesting stop after submitting batch job to SLURM'
            )
            raise ScriptEngineStopException
